/***************************************************************************
 *   Copyright (C) 2008 by Amir R. Saffari A. A.                           *
 *   amir@ymer.org                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
/**
  @author Amir Saffari <amir@ymer.org>
  @file huehist.cpp

  Extracts Robust Hue Histogram descriptors[1] from an input image and a set of key points.

  [1]. Joost van de Weijer, Cordelia Schmid, "Coloring Local Feature Extraction", ECCV 2006.
 */

#ifndef PI
#define PI 3.14159265358979323846
#endif

#include <iostream>
#include <fstream>
#include <string>
#include <blitz/array.h>

using namespace std;
using namespace blitz;

/**
 * help: prints help text.
 */
void help() {
    cout << "Extracts Robust Hue Histogram Descriptor:" << endl;
    cout << "Input arguments:" << endl;
    cout << "\t -h | --help : \t will display this message." << endl;
    cout << "\t -i : \t\t path to the input image (ppm format)." << endl;
    cout << "\t -o : \t\t path to the output file (see bellow for more information)." << endl;
    cout << "\t -k : \t\t path to the key points file (see bellow for more information)." << endl;
    cout << "\t -bin : \t number of bins of histogram (default = 36)." << endl;
    cout << "\t -smooth : \t number of smoothings (default = 2)." << endl;
    cout << "\t -lambda : \t weight of the descriptor (default = 1)." << endl;
    cout << "\t -scale : \t scale ratio (default = 1, see bellow for more information)" << endl;
    cout << "\t -sigma : \t controls the sigma for the spatial weights (default = 0.5, see bellow" << endl;
    cout << "\t\t\t for more information)." << endl << endl;
    cout << "Key points file format: " << endl;
    cout << "\t Each line is a key point in form of: x y s (anything after these three values will be" << endl;
    cout << "\t ignored till next line). 'x' and 'y' are the key point location (x is over width" << endl;
    cout << "\t and y is over height of the image). 's' is the patch size, and its value will be" << endl;
    cout << "\t multiplied by the scale ratio to obtain the final patch size." << endl << endl;
    cout << "Output file format: " << endl;
    cout << "\t Each line is the key point information (x, y, s) concatenated with the extracted hue" << endl;
    cout << "\t histogram. If you do not specify the output file, the results will be written to a" << endl;
    cout << "\t file by adding .hue extension to the key points file name." << endl << endl;
    cout << "Spatial weights: " << endl;
    cout << "\t Each pixel has a weighted effect on the histogram and the weights follow a 2-D" << endl;
    cout << "\t Gaussian form centered at the location of the key point. The width of the Gaussian" << endl;
    cout << "\t function is determined as the s*sigma/2, where s is the patch size (from the key" << endl;
    cout << "\t points). Therefore, by adjusting the value of sigma, you can control how much the" << endl;
    cout << "\t descriptor is focused on the center of the patch." << endl << endl;
    cout << "Examples:" << endl;
    cout << "\t huehist -i test.ppm -k test.key -o test.hue -bin 24" << endl;
    cout << "\t huehist -i test.ppm -k test.key -smooth 4 -lambda 0.6 -scale 10.0 -sigma 0.25" << endl << endl;
    cout << "Amir Saffari, amir@ymer.org" << endl;
    cout << "Version: 0.1" << endl;
    cout << "Reference: [1]. J. van de Weijer, C. Schmid, 'Coloring Local Feature Extraction', ECCV 2006." << endl << endl;
}

/**
 * exiterror: prints the error message and exits.
 * @param inerror: error message
 */
void exiterror(const string inerror) {
  cerr << inerror << endl;
  exit(EXIT_FAILURE);
}

/**
 * readppm: reads a ppm image file and seperates its color channels into R, G, B matrixes.
 * @param inimagefile: path to the image file
 * @return R, G, B: R-G-B channel matrixes
 */
void readppm(const string inimagefile, Array<double, 2> &outR, Array<double, 2> &outG, Array<double, 2> &outB) {
  // Checking the existance of the image
  ifstream imagefile(inimagefile.c_str( ), ios::binary);
  if ( !imagefile ) {
    exiterror("Could not access to image file: " + inimagefile);
  }
  
  // Reading the header
  int width, height, maxval;
  char tmpchar;
  string tmpstr;
  imagefile >> tmpchar;
  if (tmpchar != 'P') {
    exiterror("Image file is not in PPM format.");
  }
  imagefile >> tmpchar;
  if (tmpchar != '6') {
    exiterror("Image file is not in PPM format.");
  }
  imagefile >> tmpchar;
  if (tmpchar == '#') {
    getline(imagefile, tmpstr);
  }
  else {
    imagefile.putback(tmpchar);
  }
  imagefile >> width >> height >> maxval;
  if (width == 0 || height == 0 || maxval > 255) {
    exiterror("Image file is not in PPM format.");
  }
  
  // Reading the image data
  outR.resize(width, height);
  outG.resize(width, height);
  outB.resize(width, height);
  char *tmpline = new char [width*3];
  imagefile.get(tmpchar);
  
  for (register int ncol = 0; ncol < height; ncol++) {
    imagefile.read(tmpline, width*3);
    for (register int nrow = 0; nrow < width; nrow++) {
      outR(nrow, ncol) = (double) (tmpline[nrow*3] >= 0 ? tmpline[nrow*3] : tmpline[nrow*3] + 256);
      outG(nrow, ncol) = (double) (tmpline[nrow*3 + 1] >= 0 ? tmpline[nrow*3 + 1] : tmpline[nrow*3 + 1] + 256);
      outB(nrow, ncol) = (double) (tmpline[nrow*3 + 2] >= 0 ? tmpline[nrow*3 + 2] : tmpline[nrow*3 + 2] + 256);
    }
  }
  imagefile.close( );
  delete [] tmpline;
}

/**
 * readkeys: reads a key points file into a matrix.
 * @param inkeysfile: path to the key points file
 * @return outkeypoints: output keypoints matrix
 */
void readkeys(const string inkeysfile, Array<double, 2> &outkeypoints) {
  // Checking the existance of the key points file
  ifstream keysfile(inkeysfile.c_str( ), ios::binary);
  if ( !keysfile ) {
    exiterror("Could not access to image file: " + inkeysfile);
  }
  
  // Initial memory allocation for key points
  string tmpstr;
  int keycount = 0;
  outkeypoints.resize(10000, 3);
  while (!keysfile.eof()) {
    keysfile >> outkeypoints(keycount, 0) >> outkeypoints(keycount, 1) >> outkeypoints(keycount, 2);
    keycount++;
    if (keycount == outkeypoints.rows()) {
      outkeypoints.resizeAndPreserve(outkeypoints.rows() + 1000, 3);
    }
    if (keysfile.peek() != '\n') {
      getline(keysfile, tmpstr);
    }
  }
  outkeypoints.resizeAndPreserve(keycount - 1, 3);
  keysfile.close();
}

/**
 * write: writes a matrix to a file.
 * @param infile: path to the output file
 * @param inmat: input 2-D matrix
 */
void write(const string infile, const Array<double, 2> inmat) {
  ofstream outfile(infile.c_str(), ios::binary);
  if (!outfile) {
    exiterror("Could not open the output file for writing: " + infile);
  }
  for (int nrow = 0; nrow < inmat.rows(); nrow++) {
    for (int ncol = 0; ncol < inmat.cols(); ncol++) {
      outfile << inmat(nrow, ncol) << " ";
    }
    outfile << endl;
  }
  outfile.close();
}

/**
 * write: writes the key points and the histograms to a file.
 * @param infile: path to the output file
 * @param inhist: histograms
 * @param inkeys: key points
 */
void write(const string infile, const Array<double, 2> inhist, const Array<double, 2> inkeys) {
  ofstream outfile(infile.c_str(), ios::binary);
  if (!outfile) {
    exiterror("Could not open the output file for writing: " + infile);
  }
  for (int nrow = 0; nrow < inhist.rows(); nrow++) {
    outfile << inkeys(nrow, 0) << " " << inkeys(nrow, 1) << " " << inkeys(nrow, 2) << " ";
    for (int ncol = 0; ncol < inhist.cols(); ncol++) {
      outfile << inhist(nrow, ncol) << " ";
    }
    outfile << endl;
  }
  outfile.close();
}

/**
 * gaussian: creates a spatial Gaussian map.
 * @param inpatchsize: size of the map
 * @param insigma: sigma factor for the Gaussian map
 * @return a Gaussian map
 */
Array<double, 2> gaussian(int inpatchsize, double insigma) {
  Array<double, 2> spatweights(inpatchsize, inpatchsize);
  firstIndex i;
  secondIndex j;
  double sigma = insigma*inpatchsize/2;
  spatweights = exp(-(pow2(i - inpatchsize/2) + pow2(j - inpatchsize/2))/(2*pow2(sigma)));
  
  return spatweights;
}

/**
 * hist: computes the hue histogram.
 * @param inhue: input hue (patch)
 * @param insat: input sat (patch)
 * @param inbinnum: number of bins
 * @param insigma: sigma factor of Gaussian weights
 * @return output histogram
 */
Array<double, 1> hist(const Array<double, 2> inhue, const Array<double, 2> insat, const int inbinnum, const double insigma) {
  Array<double, 1> outhist(inbinnum);
  Array<double, 2> spatweights(insat.rows(), insat.rows());
  spatweights = gaussian(insat.rows(), insigma);
  
  int tmpind;
  outhist = 0;
  for ( register int nrow = 0; nrow < inhue.rows(); nrow++) {
    for ( register int ncol = 0; ncol < inhue.cols(); ncol++) {
      tmpind = floor(inhue(nrow, ncol)*inbinnum/(2*PI));
      outhist(tmpind) += insat(nrow, ncol)*spatweights(nrow, ncol);
    }
  }
  
  return outhist;
}

/**
 * greyenergy: computes the weighted grey scale energy.
 * @param inR: R (G, B) of the patch
 * @return output energy
 */
double greyenergy(const Array<double, 2> inR, const Array<double, 2> inG, const Array<double, 2> inB, const double insigma) {
  Array<double, 2> spatweights(inR.rows(), inR.rows());
  spatweights = gaussian(inR.rows(), insigma);
  
  return sum((inR + inG + inB)*spatweights);
}

/**
 * clshift: circular left shift of an array.
 * @param inarray: input array
 * @return left shifted array
 */
Array<double, 1> clshift(const Array<double, 1> inarray) {
  int arraysize = inarray.rows();
  Array<double, 1> outarray(arraysize);
  outarray(Range(0, arraysize - 2)) = inarray(Range(1, arraysize - 1));
  outarray(arraysize - 1) = inarray(0);
  
  return outarray;
}

/**
 * crshift: circular rigth shift of an array.
 * @param inarray: input array
 * @return right shifted array
 */
Array<double, 1> crshift(const Array<double, 1> inarray) {
  int arraysize = inarray.rows();
  Array<double, 1> outarray(arraysize);
  outarray(Range(1, arraysize - 1)) = inarray(Range(0, arraysize - 2));
  outarray(0) = inarray(arraysize - 1);
  
  return outarray;
}

/**
 * smooth: smoothes the histogram.
 * @param inhist: input histogram
 * @param insmoothnum: number of times to perform smoothing
 * @return output histogram
 */
Array<double, 1> smooth(const Array<double, 1> inhist, const int insmoothnum) {
  int binnum = inhist.rows();
  Array<double, 1> outhist(binnum);
  outhist = inhist;
  for (register int ns = 0; ns < insmoothnum; ns++) {
    outhist = (2*outhist + clshift(outhist) + crshift(outhist))/4;
  }
  
  return outhist;
}

int main( int argc, char *argv[] ){
  //=========================
  // Reading the command line
  //=========================
  string inimagefile, inkeysfile, inresfile;
  int binnum = 36, smoothnum = 2, hasoutfile = 0, haskeys = 0, inputCounter = 1;
  double lambda = 1, scaleratio = 1.0, sigmaratio = 0.5;
  
  if (argc == 1) {
    help();
    return EXIT_SUCCESS;
  }
  while (inputCounter < argc) {
    if (!strcmp(argv[inputCounter], "-h") || !strcmp(argv[inputCounter], "--help")) {
      help();
      return EXIT_SUCCESS;
    }
    else if (!strcmp(argv[inputCounter], "-i")) {
      inimagefile = argv[++inputCounter];
    }
    else if (!strcmp(argv[inputCounter], "-o")) {
      inresfile = argv[++inputCounter];
      hasoutfile = 1;
    }
    else if (!strcmp(argv[inputCounter], "-k")) {
      inkeysfile = argv[++inputCounter];
      haskeys = 1;
    }
    else if (!strcmp(argv[inputCounter], "-bin")) {
      binnum = atoi(argv[++inputCounter]);
    }
    else if (!strcmp(argv[inputCounter], "-smooth")) {
      smoothnum = atoi(argv[++inputCounter]);
    }
    else if (!strcmp(argv[inputCounter], "-lambda")) {
      lambda = atof(argv[++inputCounter]);
    }
    else if (!strcmp(argv[inputCounter], "-scale")) {
      scaleratio = atof(argv[++inputCounter]);
    }
    else if (!strcmp(argv[inputCounter], "-sigma")) {
      sigmaratio = atof(argv[++inputCounter]);
    }
    else {
      exiterror("Unknown input argument: please try --help for more information.");
    }
    inputCounter++;
  }
  if (!haskeys) {
    exiterror("Missing key points file.");
  }
  
  //==================
  // Reading the image
  //==================
  Array<double, 2> R, G, B;
  readppm(inimagefile, R, G, B);
  int width = R.rows(), height = R.cols();
  
  //=======================
  // Reading the key points
  //=======================
  Array<double, 2> keypoints;
  readkeys(inkeysfile, keypoints);
  
  //==================================
  // Extracting the hue and saturation
  //==================================
  Array<double, 2> hue(width, height), sat(width, height);
  hue = atan2(R + G - 2*B, sqrt(3)*(R - G)) + PI;
  sat = sqrt(2*(pow2(R) + pow2(G) + pow2(B) - R*G - R*B - G*B)/3 + 0.01);
  
  //==========================
  // Extracting the histograms
  //==========================
  int keynum = keypoints.rows();
  double xloc, yloc, scale;
  Array<double, 1> genergy(keynum);
  Array<double, 2> outhist(keynum, binnum);
  for (register int nkey = 0; nkey < keynum; nkey++) {
    xloc = keypoints(nkey, 0);
    yloc = keypoints(nkey, 1);
    scale = keypoints(nkey, 2)*scaleratio/2;
    
    Range wr = Range(max((int) round(xloc - scale), 0), min((int) round(xloc + scale), width - 1));
    Range hr = Range(max((int) round(yloc - scale), 0), min((int) round(yloc + scale), height - 1));
    
    outhist(nkey, Range::all()) = hist(hue(wr, hr), sat(wr, hr), binnum, sigmaratio);
    genergy(nkey) = greyenergy(R(wr, hr), G(wr, hr), B(wr, hr), sigmaratio);
  }
  
  //=========================
  // Smoothing the histograms
  //=========================
  if (smoothnum) {
    for (register int nkey = 0; nkey < keynum; nkey++) {
      outhist(nkey, Range::all()) = smooth(outhist(nkey, Range::all()), smoothnum);
    }
  }
  
  //===========================
  // Normalizing the histograms
  //===========================
  for (register int nkey = 0; nkey < keynum; nkey++) {
    outhist(nkey, Range::all()) /= (sum(outhist(nkey, Range::all())) + genergy(nkey));
  }
  outhist *= lambda;
  //outhist = where(outhist < 1e-5, 0, outhist);
  
  //====================
  // Writing the results
  //====================
  if (hasoutfile) {
    write(inresfile, outhist, keypoints);
  }
  else {
    write(inkeysfile + ".hue", outhist, keypoints);
  }
  
  return EXIT_SUCCESS;
}
